from datetime import datetime, timezone
import bcrypt
from starlette.routing import Route
from dl2050utils.core import oget
from dl2050utils.com import send_otp
from dl2050utils.restutils import OrjsonResponse, get_required, rest_ok, rest_error, log_and_raise, db_update, mk_key, mk_jwt_token

# ####################################################################################################
# auth Module
#
# Class for users authentication
#
#   All REST serving methods must return rest_ok({...}), rest_error({...}) or raise an error through log_and_raise if something wrong
#
# Main methods:
#     get_routes
#     startup
#     shutdown
#     auth_info
#     check_auth
#     check_user
#     login
#     register
#     recover_passwd
#     check_otp
#     set_passwd
#     change_passwd
#     get_users
#     update_user
#     reset_user
#     suspend_user
#     change_org
#
#
# ####################################################################################################

# ####################################################################################################
# Defenitions and helper functions
# ####################################################################################################

# 'User already registered'

NOT_REGISTERED = 0
WAIT_SET_PASSWD = 1
SUSPENDED = 2
SUSPENDED_LICENCE = 3
SUSPENDED_FAILS = 4
SUSPENDED_REQUESTS = 5
ACTIVE = 10

AUTHORIZED = 100
NOT_AUTHORIZED = 101
WRONG_PASSWD = 102
REGISTER_NOT_ALLOWED = 103
USER_ALREADY_REGISTERED = 109
OTP_ERROR = 104
OTP_EXPIRED = 105
OTP_OK = 106
PASSWORD_CHANGED = 107
UNKNOWN_USER = 108

MAX_FAILS = 5
MAX_REQUESTS = 1e6

def increase(d, f, reset=False):
    if d is None or f not in d: return
    if d[f] is None or reset: d[f] = 0
    d[f] += 1
    d['ts_access'] = datetime.now(timezone.utc)
    if f=='logins': d['ts_login'] = datetime.now(timezone.utc)

def get_age_in_seconds(ds):
    s1 = int(datetime.now(timezone.utc).strftime("%s"))
    s2 = int(datetime.strptime(ds,'%Y-%m-%d %H:%M:%S').strftime("%s"))
    return s1-s2

def get_user_info(user):
    return {
        'uid': user['id'],
        'email': user['email'],
        'status': user['ustatus'],
        'role': user['role'],
        'org': user['org'],
        'dep': user['org']
    }

def check_passwd_strengh(passwd):
    if type(passwd)!=str: return False
    if len(passwd)<6: return False
    return True

# ####################################################################################################
# Auth class
# ####################################################################################################

class Auth():
    def __init__(self, cfg, LOG, NOTIFY, db):
        self.cfg,self.LOG,self.NOTIFY,self.db = cfg,LOG,NOTIFY,db

    # ####################################################################################################
    # Helper methods
    # ####################################################################################################

    async def enforce_find_user(self, d, label2='', msg='User not found'):
        user = await self.db.select_one('users', d)
        if user is None:
            log_and_raise(self.LOG, label2=label2, msg=msg)
        return user

    async def enforce_admin_perm(self, request, label2='', msg='Not authorized', requires_su=False):
        user = await self.enforce_find_user({'id':request.user.display_name}, label2=label2, msg='Admin user not found')
        role,org = user['role'],user['org']
        if role not in ['su', 'admin'] or (requires_su==True and role!='su'):
            log_and_raise(self.LOG, label2=label2, msg=msg)
        return role,org

    async def increase_fails(self, user):
        increase(user, 'fails')
        if user['fails']>MAX_FAILS:
            user['ustatus'] = SUSPENDED_FAILS
            self.LOG(4, 0, label='AUTH', label2='login', msg={'user':user['email'], 'msg':'Too many login fails - user suspended'})
        await db_update(self.db, 'users', 'email', user)
        if user['fails'] > MAX_FAILS:
            return SUSPENDED_FAILS
        return 0

    async def register_requests(self, url, uid, t, error):
        await self.db.insert('metrics', {'ts': datetime.now(timezone.utc), 'uid':uid, 'url':url, 't':t, 'error':error})
        if uid==-1:
            return
        user = await self.db.select_one('users', {'id': uid})
        increase(user, 'requests')
        if user['requests'] > MAX_REQUESTS:
            user['ustatus'] = SUSPENDED_REQUESTS
        await db_update(self.db, 'users', 'email', user)

    async def gen_otp(self, user, return_code=WAIT_SET_PASSWD):
        if self.otp_mode=='phone':
            if oget(user,['phone']) is None:
                return rest_error(self.LOG, 'AUTH', 'register', 'phone not defined')
            user['email'] = ''
        else:
            if oget(user,['email']) is None:
                return rest_error(self.LOG, 'AUTH', 'register', 'email not defined')
            user['phone'] = ''
        otp = mk_key(n=4)
        self.LOG(2, 0, label='GEN_OTP', label2='New OTP', msg={'email': user['email'], 'OTP': otp})
        err = send_otp(self.NOTIFY, self.otp_mode, self.product, user['email'], user['phone'], otp)
        if err is not None:
            self.LOG(4, 0, label='AUTH', label2='gen_otp', msg=err)
            return rest_error(self.LOG, 'AUTH', 'register', 'send key by email error')
        user['otp'],user['ts_passwd'],user['ustatus'] =  otp, datetime.now(timezone.utc),WAIT_SET_PASSWD
        await db_update(self.db, 'users', 'email', user, prefix='gen_otp')
        return rest_ok({'user_status': return_code})

    # ####################################################################################################
    # Public methods
    # ####################################################################################################

    def get_routes(self):
        return [
            Route('/api/auth/auth_info', endpoint=self.auth_info, methods=['GET']),
            Route('/api/auth/check_user', endpoint=self.check_user, methods=['POST']),
            Route('/api/auth/login', endpoint=self.login, methods=['POST']),
            Route('/api/auth/check_otp', endpoint=self.check_otp, methods=['POST']),
            Route('/api/auth/set_passwd', endpoint=self.set_passwd, methods=['POST']),
            Route('/api/auth/recover_passwd', endpoint=self.recover_passwd, methods=['POST']),
            Route('/api/auth/change_passwd', endpoint=self.change_passwd, methods=['POST']),
            Route('/api/auth/get_users', endpoint=self.get_users, methods=['POST']),
            Route('/api/auth/register', endpoint=self.register, methods=['POST']),
            Route('/api/auth/update_user', endpoint=self.update_user, methods=['POST']),
            Route('/api/auth/reset_user', endpoint=self.reset_user, methods=['POST']),
            Route('/api/auth/suspend_user', endpoint=self.suspend_user, methods=['POST']),
            Route('/api/auth/change_org', endpoint=self.change_org, methods=['POST']),
        ]

    async def startup(self):
        try:
            self.product = self.cfg['app']['product']
            self.auth_secret = self.cfg['rest']['auth_secret']
            self.allow_register = oget(self.cfg, ['rest', 'allow_register'], False)
            self.strong_passwd = oget(self.cfg, ['rest', 'strong_passwd'], False)
            self.otp_timeout = oget(self.cfg, ['rest', 'otp_timeout'], False)
            self.otp_mode = oget(self.cfg, ['rest', 'otp_mode'], False)
        except Exception as e:
            self.LOG(4, 0, label='AUTH', label2='startup', msg={'error_msg': str(e)})
            return True
        self.LOG(2, 0, label='AUTH', label2='startup', msg=f'OK (connected to DB:{self.db.dbname})')
        return False

    async def shutdown(self):
        self.LOG(2, 0, label='AUTH', label2='shutdown', msg='OK')
        return False

    async def check_auth(self, request):
        """ 
            Method called from restapp to confirm that user token is valid.
            Valid is checked through the valid decription of request.user.display_name
        """
        user = await self.enforce_find_user({'id':request.user.display_name}, label2='check_auth')
        return get_user_info(user)

    async def auth_info(self, request):
        return rest_ok({
            'allow_register': self.allow_register,
            'strong_passwd': self.strong_passwd,
            'otp_timeout': self.otp_timeout,
            'otp_mode': self.otp_mode
        })

    async def check_user(self, request):
        """ Checks user by email prior to ask for password """
        data = await request.json()
        [email] = get_required(data,['email'])
        user = await self.db.select_one('users', {'email':email})
        if user is None:
            return rest_ok({'user_status':NOT_REGISTERED})
        if not self.allow_register and user['ustatus']==NOT_REGISTERED:
            return await self.gen_otp(user, WAIT_SET_PASSWD)
        if user['ustatus']==ACTIVE:
            increase(user, 'logins')
            await db_update(self.db, 'users', 'email', user)
            return rest_ok({'user_status':ACTIVE})
        return rest_ok({'user_status':user['ustatus']})

    async def check_otp(self, request):
        data = await request.json()
        [email,otp] = get_required(data,['email','otp'])
        user = await self.db.select_one('users', {'email': email})
        if user is None or user['ustatus']!=WAIT_SET_PASSWD or otp!=user['otp']:
            self.LOG(3, 0, label='AUTH', label2='check_otp', msg={'user': email, 'error_message': 'OTP error'})
            if await self.increase_fails(user):
                return rest_ok({'user_status': SUSPENDED_FAILS})                
            return rest_ok({'user_status': OTP_ERROR})
        delta = get_age_in_seconds(user['ts_passwd'])
        if delta > self.otp_timeout:
            self.LOG(3, 0, label='AUTH', label2='OTP', msg=f'OTP expired (delta={delta})')
            return await self.gen_otp(user, OTP_EXPIRED)
        return rest_ok({'user_status': OTP_OK})

    async def login(self, request):
        data = await request.json()
        [email,passwd] = get_required(data,['email', 'passwd'])
        user = await self.db.select_one('users', {'email':email})
        if user is None:
            self.LOG(3, 0, label='AUTH', label2='login', msg={'user':email, 'error_message':'user not registered'})
            return rest_ok({'user_status':NOT_REGISTERED})
        if user['ustatus'] in [NOT_REGISTERED, WAIT_SET_PASSWD, SUSPENDED, SUSPENDED_LICENCE]:
            self.LOG(3, 0, label='AUTH', label2='login', msg={'user':email, 'error_message': 'user not active'})
            return rest_ok({'user_status':user['ustatus']})
        if not "passwd" in user or user['passwd'] is None or passwd is None:
            return rest_error(self.LOG, 'AUTH', 'login', 'no passwd provided')
        if not bcrypt.checkpw(passwd.encode(), user['passwd'].encode()):
            self.LOG(3, 0, label='AUTH', label2='login', msg={'user': email, 'error_message': 'wrong password'})
            if await self.increase_fails(user):
                return rest_ok({'user_status': SUSPENDED_FAILS})                
            return rest_ok({'user_status': WRONG_PASSWD})
        jwt_token = mk_jwt_token(user['id'], email, self.auth_secret)
        increase(user, 'logins', True)
        user['fails'] = 0
        await db_update(self.db, 'users', 'email', user)
        return rest_ok({'user_status': AUTHORIZED, 'jwt_token': jwt_token, 'user_info': get_user_info(user)})

    async def recover_passwd(self, request):
        data = await request.json()
        [email] = get_required(data,['email'])
        user = await self.enforce_find_user({'email':email}, label2='recover_passwd')
        user['ustatus'] = WAIT_SET_PASSWD
        await db_update(self.db, 'users', 'email', user)
        await self.gen_otp(user, WAIT_SET_PASSWD)
        return rest_ok({'user_status': WAIT_SET_PASSWD})

    async def set_passwd(self, request):
        data = await request.json()
        [email,otp,passwd] = get_required(data,['email','otp','passwd'])
        user = await self.enforce_find_user({'email':email}, label2='set_passwd')
        if user['ustatus']!=WAIT_SET_PASSWD or otp!=user['otp']:
            self.LOG(3, 0, label='AUTH', label2='set_passwd', msg={'user': email, 'error_message': 'OTP error'})
            if await self.increase_fails(user):
                return rest_ok({'user_status': SUSPENDED_FAILS})                
            return rest_ok({'user_status': OTP_ERROR})
        delta = get_age_in_seconds(user['ts_passwd'])
        if delta > self.otp_timeout:
            self.LOG(3, 0, label='AUTH', label2='OTP', msg=f'OTP expired (delta={delta})')
            return await self.gen_otp(user, OTP_EXPIRED)
        hashed = bcrypt.hashpw(passwd.encode(), bcrypt.gensalt())
        user['passwd'],user['ts_passwd'],user['otp'],user['ustatus'] = hashed.decode(),datetime.now(timezone.utc),'',ACTIVE
        increase(user, 'logins')
        await db_update(self.db, 'users', 'email', user)
        jwt_token = mk_jwt_token(user['id'], user['email'], self.auth_secret)
        return rest_ok({'user_status': AUTHORIZED, 'jwt_token': jwt_token})

    async def register(self, request):
        org = None
        data = await request.json()
        if not self.allow_register:
            role,org = await self.enforce_admin_perm(request, label2='register')
            if role=='admin': org = user['org']
        [email] = get_required(data,['email'])
        user = await self.db.select_one('users', {'email':email})
        if user is not None:
            return rest_ok({'user_status': USER_ALREADY_REGISTERED})
        user = {
            'email': email,
            'ustatus': NOT_REGISTERED,
            'role': 'user',
            'org' :org,
            'ts_insert': datetime.now(timezone.utc),
            'logins' :0,
            'fails': 0,
            'requests': 0
        }
        proceed_to_set_passwd = self.allow_register and user['role']=='user'
        user['ustatus'] = WAIT_SET_PASSWD if proceed_to_set_passwd else NOT_REGISTERED
        await self.db.insert('users', user)
        if proceed_to_set_passwd:
            return await self.gen_otp(user, WAIT_SET_PASSWD)
        return rest_ok({'status': 'OK'})

    async def change_passwd(self, request):
        user = await self.enforce_find_user({'id':request.user.display_name}, label2='change_passwd')
        data = await request.json()
        [passwd,new_passwd] = get_required(data,['passwd','new_passwd'])
        if not check_passwd_strengh(passwd):
            return rest_error(self.LOG, 'AUTH', 'SERVICE_ERROR', 'change_passwd: invalid password')
        if not bcrypt.checkpw(passwd.encode(), user['passwd'].encode()):
            return rest_ok({'user_status': WRONG_PASSWD})
        hashed = bcrypt.hashpw(new_passwd.encode(), bcrypt.gensalt())
        user['passwd'],user['otp'],user['ustatus'] = hashed.decode(),'',ACTIVE
        await db_update(self.db, 'users', 'email', user)
        return rest_ok({'user_status': PASSWORD_CHANGED})

    async def get_users(self, request):
        role,org = await self.enforce_admin_perm(request, label2='get_users')
        data = await request.json()
        QARGS = ['cols','filters','sfilters','sort','ascending','offset','limit']
        kwargs = {k:data[k] for k in data if k in QARGS}
        filters =  kwargs['filters'] if 'filters' in kwargs and kwargs['filters'] is not None else []
        if role=='admin': filters.append({'col':'org', 'val':org})
        kwargs['filters'] = filters
        res = await self.db.select('users', **kwargs)
        if res is None or not 'data' in res or not 'nrows' in res:
            log_and_raise(self.LOG, label2='get_users', msg='DB access error')
        return OrjsonResponse({'status': 'OK', 'data':res['data'], 'nrows':res['nrows']})

    async def update_user(self, request):
        role,org = await self.enforce_admin_perm(request, label2='update_user')
        data = await request.json()
        user = await self.enforce_find_user({'id': data['id']}, label2='update_user')
        if role=='admin' and user['org']!=org:
            return rest_ok({'status': 'NOT AUTHORIZED'})
        await db_update(self.db, 'users', 'id', data, 'update_user')
        return rest_ok({'status': 'OK'})

    async def reset_user(self, request):
        role,org = await self.enforce_admin_perm(request, label2='reset_user')
        data = await request.json()
        user = await self.enforce_find_user({'id': data['id']}, label2='reset_user')
        if role=='admin' and user['org']!=org:
            return rest_ok({'status': 'NOT AUTHORIZED'})
        user['ustatus'],user['fails'],user['otp'] = 0,0,None
        await db_update(self.db, 'users', 'id', user, 'reset_user')
        return rest_ok({'status': 'OK'})

    async def suspend_user(self, request):
        """ Suspends a user changing ustatus to SUSPENDED or to SUSPENDED_LICENCE if data['ustatus']==SUSPENDED_LICENCE """
        role,org = await self.enforce_admin_perm(request, label2='suspend_user')
        data = await request.json()
        user = await self.enforce_find_user({'id': data['id']}, label2='suspend_user')
        if role=='admin' and user['org']!=org:
            return rest_ok({'status': 'NOT AUTHORIZED'})
        ustatus = int(data['ustatus']) if 'ustatus' in data and int(data['ustatus'])==SUSPENDED_LICENCE else SUSPENDED
        user['ustatus'] = ustatus
        await db_update(self.db, 'users', 'id', user, 'reset_user')
        return rest_ok({'status': 'OK'})

    async def change_org(self, request):
        _,_ = await self.enforce_admin_perm(request, label2='change_org', requires_su=True)
        user = await self.enforce_find_user({'id':request.user.display_name}, label2='change_org')
        data = await request.json()
        user['org'] = data['org']
        jwt_token = mk_jwt_token(user['id'], user['email'], self.auth_secret)
        _ = await self.db.update('users', 'id', user)
        return rest_ok({'status': 'OK', 'jwt_token': jwt_token, 'user_info': get_user_info(user)})
